// ====================================================================
// Copyright (C) 2011 - NII - Benoit Goepfert
// Copyright (C) 2005 - INRIA - Sage Group (IRISA)
// Copyright (C) 2011 - DIGITEO - Michael Baudin
// 
// This file must be used under the terms of the CeCILL.
// This source file is licensed as described in the file COPYING, which
// you should have received as part of this distribution.  The terms
// are also available at    
// http://www.cecill.info/licences/Licence_CeCILL_V2-en.txt
// ====================================================================

// From Scilab
#include "api_scilab.h"
#include "stack-c.h"
#include "Scierror.h"

// From the lib
#include "conv.h"

// From ANSI C
#include "stdlib.h"
#include "spilu.h"

// From this gateway
#include "gw_spilu_support.h"

extern int sci_ilu0(char *fname)
{ 
	int mA=0, nA=0, nnzLU=0;
	int *iw=NULL, *jlu=NULL, *ju=NULL, *ia=NULL;
	double *alu=NULL;
	SciSparse A,*L=NULL,*U=NULL;
	int ierr;
	int readFlag;

	CheckRhs(1,1);
	CheckLhs(2,2);
	// Argument 1: A (sparse)
	if (VarType(1)==5)
	{
		GetRhsVar(1,"s",&mA,&nA,&A);    
		if (mA!=nA)
		{
			Scierror(501,"%s: input matrix must be square \r\n",fname);
			return(0);
		}
	}
	else
	{
		Scierror(501,"%s: input matrix must be sparse \r\n",fname);
		return(0);
	}

	//
	// Proceed...
	readFlag = spilu_allocRowptr(fname, A, &ia);
	if (readFlag==0)
	{
		return 0;
	}
	nnzLU= A.nel;
	readFlag = spilu_allocLUArrays(fname, nnzLU, A.m, &alu, &jlu, &ju);
	if (readFlag==0)
	{
		return 0;
	}
	iw= (int *) malloc((A.m)*sizeof(int));
	if (iw==NULL)
	{
		Scierror(112, "%s: No more memory.\n",fname);
		return 0;
	}

	F2C(ilu0)(&A.m,A.R,A.icol,ia,alu,jlu,ju,iw,&ierr);

	free(iw);
	free(ia);

	//
	// Error case
	if (ierr!=0){
		free(ju);
		free(jlu);
		free(alu);    
		Scierror(501,"%s: zero pivot encountered at step number %d \r\n",fname,ierr);
		return(0);
	}

	//
	// Regular case
	spiluc_spluget(A.m,ju,jlu,alu,&L,&U);
	free(ju);
	free(jlu);
	free(alu);
	CreateVarFromPtr(Rhs+1,"s",&L->m,&L->n,L);
	CreateVarFromPtr(Rhs+2,"s",&U->m,&U->n,U);
	LhsVar(1) = Rhs+1;
	LhsVar(2) = Rhs+2;

	return(0);
}

