// ====================================================================
// Copyright (C) 2011 - NII - Benoit Goepfert
// Copyright (C) 2005 - INRIA - Sage Group (IRISA)
// Copyright (C) 2011 - DIGITEO - Michael Baudin
// 
// This file must be used under the terms of the CeCILL.
// This source file is licensed as described in the file COPYING, which
// you should have received as part of this distribution.  The terms
// are also available at    
// http://www.cecill.info/licences/Licence_CeCILL_V2-en.txt
// ====================================================================

// From Scilab
#include "api_scilab.h"
#include "stack-c.h"
#include "Scierror.h"

// From the lib
#include "conv.h"
#include "spilu.h"

// From ANSI C
#include "stdlib.h"

// From this gateway
#include "gw_spilu_support.h"


extern int sci_iludp(char *fname)
{ 
	int mA=0, nA=0, i=0;
	int mperm=1, nperm=0, pperm=0;
	int iwk=0;
	int *jw=NULL, *jlu=NULL, *ju=NULL, *vperm=NULL, *ia=NULL;
	double *w=NULL, *alu=NULL;
	SciSparse A, *L=NULL, *U=NULL;
	int ierr;
	double alpha;
	double drop;
	double permtol;
	int bloc;
	int readFlag;
	int nnzLU;

	CheckRhs(1,5);
	CheckLhs(3,3);
	// Argument 1: A (sparse)
	if (VarType(1)==5){
		GetRhsVar(1,"s",&mA,&nA,&A);
		if (mA!=nA){
			Scierror(501,"%s: input matrix must be square \r\n",fname);
			return(0);
		}
	} else {
		Scierror(501,"%s: input matrix must be sparse \r\n",fname);
		return 0;
	}
	// Argument 2: alph (double) positive
	readFlag = spilu_getArgumentAlpha ( fname , 2, Rhs, &alpha );
	if (readFlag==0)
	{
		return 0;
	}
	// Argument 3: drop (double) positive
	readFlag = spilu_getArgumentDrop ( fname , 3, Rhs, A , &drop );
	if (readFlag==0)
	{
		return 0;
	}
	// Argument 4: permtol (double) positive
	readFlag = spilu_getArgumentPermtol ( fname, 4, Rhs, &permtol );
	if (readFlag==0)
	{
		return 0;
	}
	// Argument 5: bloc (double) positive, integer value
	readFlag = spilu_getArgumentBloc ( fname , 5, Rhs, A, &bloc );
	if (readFlag==0)
	{
		return 0;
	}
	
	//
	// Proceed...
	nperm=A.m;
	CreateVar(Rhs+1,"i",&mperm,&nperm,&pperm);

	readFlag = spilu_allocRowptr(fname, A, &ia);
	if (readFlag==0)
	{
		return 0;
	}
	nnzLU = spiluc_lband(A)+A.m+1;
	readFlag = spilu_allocLUArrays(fname, nnzLU, A.m, &alu, &jlu, &ju);
	if (readFlag==0)
	{
		return 0;
	}

	iwk = nnzLU + 1;
	jw=(int *) malloc(2*A.m*sizeof(int));
	if (jw==NULL)
	{
		Scierror(112, "%s: No more memory.\n",fname);
		return 0;
	}
	w= (double *) malloc(2*A.m*sizeof(double));
	if (w==NULL)
	{
		Scierror(112, "%s: No more memory.\n",fname);
		return 0;
	}
	vperm= (int *) malloc(2*A.m*sizeof(int));
	if (vperm==NULL)
	{
		Scierror(112, "%s: No more memory.\n",fname);
		return 0;
	}

	F2C(iludp)(&A.m,A.R,A.icol,ia,&alpha,&drop,&permtol,&bloc,alu,jlu,ju,&iwk,w,jw,vperm,&ierr);

	free(w);
	free(jw);
	free(ia);

	//
	// Error case
	if (ierr!=0){
		free(ju);
		free(jlu);
		free(alu);
		free(vperm);
		if (ierr==-1) {
			Scierror(501,"%s: input matrix may be wrong \r\n",fname);
			return(0);
		} else if (ierr==-2) {
			Scierror(501,"%s: not enough memory for matrix L or U \r\n",fname);
			return(0);
		} else if (ierr==-3) {
			Scierror(501,"%s: zero row encountered \r\n",fname);
			return(0);
		} else {
			Scierror(501,"%s: zero pivot encountered at step number %d \r\n",fname,ierr);
			return(0);
		}
	}
	//
	// Regular case
	for (i=0;i<A.m;i++) {
		istk(pperm)[i]=vperm[i];
	}
	free(vperm);

	spiluc_spluget(A.m,ju,jlu,alu,&L,&U);
	free(ju);
	free(jlu);
	free(alu);
	CreateVarFromPtr(Rhs+2,"s",&L->m,&L->n,L);
	CreateVarFromPtr(Rhs+3,"s",&U->m,&U->n,U);

	LhsVar(1) = Rhs+2;
	LhsVar(2) = Rhs+3;
	LhsVar(3) = Rhs+1;
	return(0);
}
