// ====================================================================
// Copyright (C) 2011 - NII - Benoit Goepfert
// Copyright (C) 2005 - INRIA - Sage Group (IRISA)
// Copyright (C) 2011 - DIGITEO - Michael Baudin
// 
// This file must be used under the terms of the CeCILL.
// This source file is licensed as described in the file COPYING, which
// you should have received as part of this distribution.  The terms
// are also available at    
// http://www.cecill.info/licences/Licence_CeCILL_V2-en.txt
// ====================================================================

// From Scilab
#include "api_scilab.h"
#include "stack-c.h"
#include "Scierror.h"

// From the lib
#include "conv.h"
#include "spilu.h"

// From ANSI C
#include "stdlib.h"

// From this gateway
#include "gw_spilu_support.h"

extern int sci_ilut(char *fname)
{ 
	int mA=0, iwk=0, nA=0;
	int *jw=NULL, *ia=NULL, *jlu=NULL, *ju=NULL;
	double *w=NULL,*alu=NULL;
	SciSparse A,*L=NULL,*U=NULL;
	int ierr;
	int lfil;
	double drop;
	int readFlag;
	int nnzLU;

	CheckRhs(1,3);
	CheckLhs(2,2);
	// Argument 1: A (sparse)
	if (VarType(1)==5){
		GetRhsVar(1,"s",&mA,&nA,&A);
		if (mA!=nA){
			Scierror(501,"%s: input matrix must be square \r\n",fname);
			return(0);
		}
	} else {
		Scierror(501,"%s: input matrix must be sparse \r\n",fname);
		return 0;
	}
	// Argument 2: lfil (double) positive, integer value
	readFlag= spilu_getArgumentLfil ( fname, 2, Rhs, A, &lfil );
	if (readFlag==0)
	{
		return 0;
	}
	// Argument 3: drop (double) positive
	readFlag = spilu_getArgumentDrop ( fname, 3, Rhs, A, &drop );
	if (readFlag==0)
	{
		return 0;
	}

	//
	// Proceed...
	readFlag = spilu_allocRowptr(fname, A, &ia);
	if (readFlag==0)
	{
		return 0;
	}
	nnzLU = (A.m)*(2*lfil+1);
	readFlag = spilu_allocLUArrays(fname, nnzLU, A.m, &alu, &jlu, &ju);
	if (readFlag==0)
	{
		return 0;
	}

	iwk = nnzLU + 1;
	jw=(int *) malloc(2*A.m*sizeof(int));
	if (jw==NULL)
	{
		Scierror(112, "%s: No more memory.\n",fname);
		return 0;
	}
	w= (double *) malloc(A.m*sizeof(double));
	if (w==NULL)
	{
		Scierror(112, "%s: No more memory.\n",fname);
		return 0;
	}

	F2C(ilut)(&A.m,A.R,A.icol,ia,&lfil,&drop,alu,jlu,ju,&iwk,w,jw,&ierr);
	free(w);
	free(jw);
	free(ia);

	//
	// Error case
	if (ierr!=0){
		free(ju);
		free(jlu);
		free(alu);
		if (ierr==-1) {
			Scierror(501,"%s: input matrix may be wrong \r\n",fname);
			return(0);
		} else if (ierr==-2) {
			Scierror(501,"%s: not enough memory for matrix L \r\n",fname);
			return(0);
		} else if (ierr==-3) {
			Scierror(501,"%s: not enough memory for matrix U \r\n",fname);
			return(0);
		} else if (ierr==-4) {
			Scierror(501,"%s: illegal value for lfil \r\n",fname);
			return(0);
		} else if (ierr==-5) {
			Scierror(501,"%s: zero row encountered in A or U \r\n",fname);
			return(0);
		} else {
			Scierror(501,"%s: zero pivot encountered at step number %d \r\n",fname,ierr);
			return(0);
		}
	}
	//
	// Regular case
	spiluc_spluget(A.m,ju,jlu,alu,&L,&U);
	free(ju);
	free(jlu);
	free(alu);
	CreateVarFromPtr(Rhs+1,"s",&L->m,&L->n,L);
	CreateVarFromPtr(Rhs+2,"s",&U->m,&U->n,U);
	LhsVar(1) = Rhs+1;
	LhsVar(2) = Rhs+2;
	return(0);
}
